@@ -38,12 +38,21 @@
 
 #ifdef HAVE_NCURSES
 #include <ncurses.h>
+#ifdef HAVE_TERM_H
 #include <term.h>
+#endif
+#else
+#ifdef HAVE_CURSES
+#include <curses.h>
+#ifdef HAVE_TERM_H
+#include <term.h>
+#endif
 #else
 #ifdef HAVE_TERMCAP
 #include <termcap.h>
 #endif
 #endif
+#endif
 
 #define BUF_SIZE 512
 #define MAX_COLS 512
