@@ -16,15 +16,29 @@
  * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
  */
 
-#ident "$Id: vteapp.c,v 1.45 2002/11/05 22:54:20 nalin Exp $"
+#ident "$Id: vteapp.c,v 1.51 2003/03/05 22:31:07 nalin Exp $"
 #include "../config.h"
+#include <sys/ioctl.h>
 #include <sys/stat.h>
+#include <fcntl.h>
 #include <string.h>
 #include <unistd.h>
 #include <gtk/gtk.h>
 #include <glib-object.h>
+#ifdef HAVE_XFT2
+#include <fontconfig/fontconfig.h>
+#endif
 #include "debug.h"
 #include "vte.h"
+#include <gdk/gdkx.h>
+
+#ifdef ENABLE_NLS
+#include <libintl.h>
+#define _(String) dgettext(PACKAGE, String)
+#else
+#define _(String) String
+#define bindtextdomain(package,dir)
+#endif
 
 #define DINGUS1 "(((news|telnet|nttp|file|http|ftp|https)://)|(www|ftp)[-A-Za-z0-9]*\\.)[-A-Za-z0-9\\.]+(:[0-9]*)?"
 #define DINGUS2 "(((news|telnet|nttp|file|http|ftp|https)://)|(www|ftp)[-A-Za-z0-9]*\\.)[-A-Za-z0-9\\.]+(:[0-9]*)?/[-A-Za-z0-9_\\$\\.\\+\\!\\*\\(\\),;:@&=\\?/~\\#\\%]*[^]'\\.}>\\) ,\\\"]"
@@ -282,6 +296,80 @@ decrease_font_size(GtkWidget *widget, gpointer data)
 	adjust_font_size(widget, data, -1);
 }
 
+static void
+mess_with_fontconfig(void)
+{
+#ifdef HAVE_XFT2
+	/* Is this even a good idea?  Probably not, since this doesn't expose
+	 * these fonts to the gnome-font-properties capplet. */
+	FcInit();
+	FcConfigAppFontAddDir(NULL, DATADIR "/" PACKAGE "/fonts");
+#endif
+}
+
+static gboolean
+read_and_feed(GIOChannel *source, GIOCondition condition, gpointer data)
+{
+	char buf[2048];
+	gsize size;
+	GIOStatus status;
+	g_return_val_if_fail(VTE_IS_TERMINAL(data), FALSE);
+	status = g_io_channel_read_chars(source, buf, sizeof(buf),
+					 &size, NULL);
+	if ((status == G_IO_STATUS_NORMAL) && (size > 0)) {
+		vte_terminal_feed(VTE_TERMINAL(data), buf, size);
+		return TRUE;
+	}
+	return FALSE;
+}
+
+static void
+disconnect_watch(GtkWidget *widget, gpointer data)
+{
+	g_source_remove(GPOINTER_TO_INT(data));
+}
+
+static void
+clipboard_get(GtkClipboard *clipboard, GtkSelectionData *selection_data,
+	      guint info, gpointer owner)
+{
+	/* No-op. */
+	return;
+}
+
+static void
+take_xconsole_ownership(GtkWidget *widget, gpointer data)
+{
+	char *name, hostname[255];
+	GdkDisplay *display;
+	Atom xatom;
+	GdkAtom atom;
+	GtkClipboard *clipboard;
+	GtkTargetEntry targets[] = {
+		{"UTF8_STRING", 0, 0},
+		{"COMPOUND_TEXT", 0, 0},
+		{"TEXT", 0, 0},
+		{"STRING", 0, 0},
+	};
+
+	memset(hostname, '\0', sizeof(hostname));
+	gethostname(hostname, sizeof(hostname) - 1);
+	display = gdk_display_get_default();
+
+	name = g_strdup_printf("MIT_CONSOLE_%s", hostname);
+	xatom = gdk_x11_get_xatom_by_name_for_display(display, name);
+	atom = gdk_x11_xatom_to_atom_for_display(display, xatom);
+	clipboard = gtk_clipboard_get(atom);
+	g_free(name);
+
+	gtk_clipboard_set_with_owner(clipboard,
+				     targets,
+				     G_N_ELEMENTS(targets),
+				     clipboard_get,
+				     (GtkClipboardClearFunc)gtk_main_quit,
+				     G_OBJECT(widget));
+}
+
 int
 main(int argc, char **argv)
 {
@@ -289,7 +377,8 @@ main(int argc, char **argv)
 	char *env_add[] = {"FOO=BAR", "BOO=BIZ", NULL};
 	const char *background = NULL;
 	gboolean transparent = FALSE, audible = TRUE, blink = TRUE,
-		 debug = FALSE, dingus = FALSE, geometry = TRUE;
+		 debug = FALSE, dingus = FALSE, geometry = TRUE, dbuffer = TRUE,
+		 console = FALSE, scroll = FALSE;
 	long lines = 100;
 	const char *message = "Launching interactive shell...\r\n";
 	const char *font = NULL;
@@ -300,10 +389,12 @@ main(int argc, char **argv)
 	int opt;
 	int i, j;
 	GList *args = NULL;
-	GdkColor fore, back;
+	GdkColor fore, back, tint;
+	GdkCursor *gumby = NULL, *hand = NULL;
 	const char *usage = "Usage: %s "
 			    "[ [-B image] | [-T] ] "
 			    "[-D] "
+			    "[-2] "
 			    "[-a] "
 			    "[-b] "
 			    "[-c command] "
@@ -315,6 +406,8 @@ main(int argc, char **argv)
 			    "[-t terminaltype]\n";
 	back.red = back.green = back.blue = 0xffff;
 	fore.red = fore.green = fore.blue = 0x0000;
+	tint.red = tint.green = tint.blue = 0;
+	tint = back;
 
 	/* Have to do this early. */
 	if (getenv("VTE_PROFILE_MEMORY")) {
@@ -342,17 +435,23 @@ main(int argc, char **argv)
 	argv2[i] = NULL;
 	g_assert(i < (g_list_length(args) + 2));
 	/* Parse some command-line options. */
-	while ((opt = getopt(argc, argv, "B:DTabc:df:ghn:t:w:")) != -1) {
+	while ((opt = getopt(argc, argv, "B:CDT2abc:df:ghn:st:w:")) != -1) {
 		switch (opt) {
 			case 'B':
 				background = optarg;
 				break;
+			case 'C':
+				console = TRUE;
+				break;
 			case 'D':
 				dingus = TRUE;
 				break;
 			case 'T':
 				transparent = TRUE;
 				break;
+			case '2':
+				dbuffer = !dbuffer;
+				break;
 			case 'a':
 				audible = !audible;
 				break;
@@ -377,6 +476,9 @@ main(int argc, char **argv)
 					lines = 100;
 				}
 				break;
+			case 's':
+				scroll = !scroll;
+				break;
 			case 't':
 				terminal = optarg;
 				break;
@@ -406,6 +508,7 @@ main(int argc, char **argv)
 
 	/* Create the terminal widget and add it to the scrolling shell. */
 	widget = vte_terminal_new();
+	gtk_widget_set_double_buffered(widget, dbuffer);
 	gtk_box_pack_start(GTK_BOX(hbox), widget, TRUE, TRUE, 0);
 
 	/* Connect to the "char_size_changed" signal to set geometry hints
@@ -468,6 +571,7 @@ main(int argc, char **argv)
 	/* Set some defaults. */
 	vte_terminal_set_audible_bell(VTE_TERMINAL(widget), audible);
 	vte_terminal_set_cursor_blinks(VTE_TERMINAL(widget), blink);
+	vte_terminal_set_scroll_background(VTE_TERMINAL(widget), scroll);
 	vte_terminal_set_scroll_on_output(VTE_TERMINAL(widget), FALSE);
 	vte_terminal_set_scroll_on_keystroke(VTE_TERMINAL(widget), TRUE);
 	vte_terminal_set_scrollback_lines(VTE_TERMINAL(widget), lines);
@@ -480,11 +584,15 @@ main(int argc, char **argv)
 		vte_terminal_set_background_transparent(VTE_TERMINAL(widget),
 							TRUE);
 	}
+	vte_terminal_set_background_tint_color(VTE_TERMINAL(widget), &tint);
 	vte_terminal_set_colors(VTE_TERMINAL(widget), &fore, &back, NULL, 0);
 	if (terminal != NULL) {
 		vte_terminal_set_emulation(VTE_TERMINAL(widget), terminal);
 	}
 
+	/* Mess with our fontconfig setup. */
+	mess_with_fontconfig();
+
 	/* Set the default font. */
 	if (font != NULL) {
 		vte_terminal_set_font_from_string(VTE_TERMINAL(widget), font);
@@ -493,22 +601,76 @@ main(int argc, char **argv)
 	/* Match "abcdefg". */
 	vte_terminal_match_add(VTE_TERMINAL(widget), "abcdefg");
 	if (dingus) {
-		vte_terminal_match_add(VTE_TERMINAL(widget), DINGUS1);
-		vte_terminal_match_add(VTE_TERMINAL(widget), DINGUS2);
+		i = vte_terminal_match_add(VTE_TERMINAL(widget), DINGUS1);
+		gumby = gdk_cursor_new(GDK_GUMBY);
+		vte_terminal_match_set_cursor(VTE_TERMINAL(widget), i, gumby);
+		gdk_cursor_unref(gumby);
+		hand = gdk_cursor_new(GDK_HAND1);
+		i = vte_terminal_match_add(VTE_TERMINAL(widget), DINGUS2);
+		vte_terminal_match_set_cursor(VTE_TERMINAL(widget), i, hand);
+		gdk_cursor_unref(hand);
 	}
 
-	/* Launch a shell. */
+	if (console) {
+		/* Open a "console" connection. */
+		int consolefd = -1, yes = 1, watch;
+		GIOChannel *channel;
+		consolefd = open("/dev/console", O_RDONLY | O_NOCTTY);
+		if (consolefd != -1) {
+			/* Assume failure. */
+			console = FALSE;
+#ifdef TIOCCONS
+			if (ioctl(consolefd, TIOCCONS, &yes) != -1) {
+				/* Set up a listener. */
+				channel = g_io_channel_unix_new(consolefd);
+				watch = g_io_add_watch(channel,
+						       G_IO_IN,
+						       read_and_feed,
+						       widget);
+				g_signal_connect(G_OBJECT(widget),
+						 "eof",
+						 G_CALLBACK(disconnect_watch),
+						 GINT_TO_POINTER(watch));
+				g_signal_connect(G_OBJECT(widget),
+						 "child-exited",
+						 G_CALLBACK(disconnect_watch),
+						 GINT_TO_POINTER(watch));
+				g_signal_connect(G_OBJECT(widget),
+						 "realize",
+						 G_CALLBACK(take_xconsole_ownership),
+						 NULL);
 #ifdef VTE_DEBUG
-	if (_vte_debug_on(VTE_DEBUG_MISC)) {
-		vte_terminal_feed(VTE_TERMINAL(widget), message,
-				  strlen(message));
-	}
+				vte_terminal_feed(VTE_TERMINAL(widget),
+						  "Console log for ...\r\n",
+						  -1);
 #endif
-	vte_terminal_fork_command(VTE_TERMINAL(widget),
-				  command, NULL, env_add, working_directory,
-				  TRUE, TRUE, TRUE);
-	if (command == NULL) {
-		vte_terminal_feed_child(VTE_TERMINAL(widget), "pwd\n", -1);
+				/* Record success. */
+				console = TRUE;
+			}
+#endif
+		} else {
+			/* Bail back to normal mode. */
+			g_warning(_("Could not open console.\n"));
+			close(consolefd);
+			console = FALSE;
+		}
+	}
+	if (!console) {
+		/* Launch a shell. */
+#ifdef VTE_DEBUG
+		if (_vte_debug_on(VTE_DEBUG_MISC)) {
+			vte_terminal_feed(VTE_TERMINAL(widget), message,
+					  strlen(message));
+		}
+#endif
+		vte_terminal_fork_command(VTE_TERMINAL(widget),
+					  command, NULL, env_add,
+					  working_directory,
+					  TRUE, TRUE, TRUE);
+		if (command == NULL) {
+			vte_terminal_feed_child(VTE_TERMINAL(widget),
+						"pwd\n", -1);
+		}
 	}
 
 	/* Go for it! */
